import micromatch from "micromatch";
import { builtinModules } from "module";
import * as path from "path";
import { createBundler } from "./bundle.js";
import * as fsutils from "./fsutils.js";
import { getEntrySrc } from "./html.js";
import { renderStaticPages, urlToFileName, } from "./prerender.js";
export function capri({ createIndexFiles = true, prerender = "/", followLinks = true, islandGlobPattern = "/src/**/*.island.*", lagoonGlobPattern = "/src/**/*.lagoon.*", templateProcessing, adapter, target, spa, }) {
    let mode;
    const ssr = resolveRelative("./virtual/ssr.js");
    let serverEntry;
    /** The project's root directory as defined in Vite config or the cwd */
    let rootDir;
    /** Absolute path of the build output dir */
    let outDir;
    /** The BASE_URL */
    let base;
    /** The content of index.html as genereated by the client build */
    let template;
    let manifest;
    const { injectWrapper = "onLoad" } = adapter;
    /**
     * Test if id matches the given glob pattern. If so, check if a wrapper
     * is provided by the adapter and set the meta data accordingly.
     */
    function resolveWrapper(id, pattern, candidates) {
        // Exclude sources that contain "?", e.g. vue styles and setup scripts
        if (!id.includes("?") && micromatch.contains(id, pattern)) {
            const wrapper = candidates[mode];
            if (wrapper) {
                return {
                    id,
                    meta: { wrapper, wrapped: id },
                };
            }
        }
    }
    /**
     * Load the wrapper as requested in the meta data.
     * In the wrapper code, two replacements are applied:
     *
     * - "virtual:capri-component" -> the wrapped module + "?unwrapped"
     * - "%COMPONENT_ID%" -> the Id of the wrapped component (relative to the root)
     */
    function loadWrapper(meta) {
        const { wrapper, wrapped } = meta;
        // Note: we add the basename so that the extension stays the same...
        const unwrappedId = wrapped + "?unwrapped=" + path.basename(wrapped);
        const componentId = wrapped.startsWith(rootDir)
            ? wrapped.slice(rootDir.length)
            : wrapped;
        const template = fsutils.read(wrapper);
        return template
            .replace(/virtual:capri-component/g, unwrappedId)
            .replace(/%COMPONENT_ID%/g, componentId);
    }
    return [
        {
            // Allow build targets to modify the config
            name: "vite-plugin-capri-target",
            // Needs to run before the main plugin
            enforce: "pre",
            config: target?.config,
        },
        {
            name: "vite-plugin-capri-main",
            // we need to modify the bundle before vite:build-html runs
            enforce: "pre",
            config(config, { command }) {
                if (config.build?.ssr) {
                    mode = "server";
                }
                else if (command === "serve") {
                    // running in dev server
                    mode = "spa";
                }
                else {
                    mode = "client";
                }
                rootDir = path.resolve(config.root ?? "");
                outDir = path.resolve(rootDir, config.build?.outDir ?? "dist");
                serverEntry = getServerEntryScript(config);
                // Allow base to be set via env:
                base = config.base ?? process.env.BASE_URL ?? "/";
                if (spa)
                    spa = urlToFileName(spa, createIndexFiles, base);
                if (mode === "server") {
                    // Read the index.html so we can use it as template for all prerendered pages.
                    const indexHtml = path.join(outDir, "index.html");
                    template = fsutils.read(indexHtml);
                    fsutils.rm(indexHtml);
                    manifest = readManifest(outDir);
                    if (!templateProcessing && target) {
                        // default to regexp when a build target is set
                        templateProcessing = "regexp";
                    }
                    return {
                        base,
                        define: {
                            USE_CHEERIO: templateProcessing === "regexp" ? "false" : "true",
                            "process.env.SSR": "true",
                        },
                        ssr: {
                            // The capri packages can't be externalized as they need to be
                            // processed by Vite (virtual modules and glob imports).
                            noExternal: ["capri", /@capri-js\//],
                        },
                        build: {
                            ssr,
                            emptyOutDir: false, // keep the client build
                        },
                    };
                }
                else {
                    // Client build ...
                    let rollupOptions = {};
                    if (isServerEntryScript(config)) {
                        // index.html points to a .server.* file
                        // This implies that the site can't be rendered as SPA.
                        if (spa) {
                            throw new Error("In order to generate an SPA, index.html must point to a client entry file.");
                        }
                        // We nevertheless want Vite to perform a client-build in order to
                        // extract all assets. As most server-side rendering solutions use
                        // node built-ins we mark them as external so that Rollup doesn't
                        // try to bundle them (which would fail).
                        rollupOptions = {
                            external: builtinModules,
                        };
                    }
                    else if (spa) {
                        // Generate two entry chunks:
                        rollupOptions = {
                            input: {
                                index: "/index.html",
                                spa,
                            },
                        };
                    }
                    return {
                        base,
                        build: {
                            ssrManifest: true,
                            rollupOptions,
                        },
                    };
                }
            },
            async resolveId(source, importer, options) {
                if (source === spa) {
                    return resolveFile(this, spa);
                }
                if (source === "virtual:capri-hydration") {
                    return { id: "\0virtual:capri-hydration", moduleSideEffects: true };
                }
                if (source === "virtual:capri-hydrate") {
                    return this.resolve(adapter.hydrate);
                }
                if (source === "virtual:capri-render") {
                    return serverEntry;
                }
                if (!source.includes("?unwrapped")) {
                    const resolved = await this.resolve(source, importer, {
                        ...options,
                        skipSelf: true,
                    });
                    if (resolved) {
                        return (resolveWrapper(resolved.id, islandGlobPattern, adapter.island) ??
                            resolveWrapper(resolved.id, lagoonGlobPattern, adapter.lagoon));
                    }
                }
            },
            async load(id) {
                if (spa && id.endsWith(spa)) {
                    const index = await resolveIndexHtml(this);
                    return fsutils.read(index);
                }
                if (id === "\0virtual:capri-hydration") {
                    const file = resolveRelative("./virtual/hydration.js");
                    return fsutils
                        .read(file)
                        .replace(/%ISLAND_GLOB_PATTERN%/g, islandGlobPattern);
                }
                if (id === ssr) {
                    return fsutils
                        .read(ssr)
                        .replace(/"%TEMPLATE%"/, JSON.stringify(template))
                        .replace("{/*MANIFEST*/}", JSON.stringify(manifest));
                }
            },
            async buildStart() {
                if (mode === "client") {
                    // Add the hydration code to the bundle
                    this.emitFile({
                        id: "virtual:capri-hydration",
                        type: "chunk",
                        name: "hydrate",
                    });
                }
            },
            async generateBundle(options, bundle) {
                if (mode === "client") {
                    // Modify the bundle so that the index.html file will load the
                    // "hydrate" rather than the "index" chunk but will still include
                    // all the original assets:
                    let indexModuleId;
                    let indexMetadata;
                    Object.values(bundle).forEach((chunk) => {
                        if (chunk.type === "chunk" &&
                            chunk.facadeModuleId &&
                            chunk.name === "index") {
                            // Remember the id and the assets of the index.html file...
                            indexModuleId = chunk.facadeModuleId;
                            indexMetadata = chunk.viteMetadata;
                            // Merge viteMetadata of imports. When code is split, assets
                            // can move further down in the tree.
                            gatherCss(chunk, bundle, indexMetadata.importedCss);
                            // Set the facadeModuleId to null so that vite:build-html does
                            // not add script tags for it:
                            chunk.facadeModuleId = null;
                        }
                        else if (chunk.type === "chunk" && chunk.name === "hydrate") {
                            // Set the facadeModuleId of the hydration code to the id of
                            // the index.html module so that vite:build-html creates a
                            // script tag for it:
                            chunk.facadeModuleId = indexModuleId;
                            chunk.viteMetadata = indexMetadata;
                        }
                    });
                }
            },
            async writeBundle(options, bundle) {
                if (mode === "server") {
                    const ssrBundle = path.resolve(outDir, "ssr.js");
                    // Prerender pages...
                    const prerendered = await renderStaticPages({
                        ssrBundle,
                        createIndexFiles,
                        outDir,
                        base,
                        prerender,
                        followLinks,
                    });
                    // Hook for build targets
                    if (target) {
                        await target.build({
                            rootDir,
                            outDir,
                            ssrBundle,
                            prerendered,
                            fsutils,
                            bundle: createBundler(ssrBundle),
                        });
                    }
                    fsutils.rm(ssrBundle);
                }
            },
        },
        injectWrapper === "onLoad"
            ? {
                name: "capri-load",
                load(id) {
                    const info = this.getModuleInfo(id);
                    if (isWrapperInfo(info))
                        return loadWrapper(info.meta);
                },
            }
            : {
                name: "capri-transform",
                enforce: "post",
                transform(code, id) {
                    const info = this.getModuleInfo(id);
                    if (isWrapperInfo(info))
                        return loadWrapper(info.meta);
                },
            },
    ];
}
function resolveRelative(src) {
    return new URL(src, import.meta.url).pathname;
}
function getEntryScript(config) {
    const indexHtml = path.resolve(config.root ?? "", "index.html");
    const src = getEntrySrc(fsutils.read(indexHtml));
    if (!src)
        throw new Error(`Can't find entry script in ${indexHtml}`);
    return path.join(path.resolve(config.root ?? ""), src);
}
function getServerEntryScript(config) {
    const clientEntryScript = getEntryScript(config);
    const f = clientEntryScript.replace(/(\.client|\.server)?(\.[^.]+)$/, ".server$2");
    if (!fsutils.exists(f)) {
        throw new Error(`File not found: ${f}. Make sure to name your server entry file accordingly.`);
    }
    return f;
}
export function isServerEntryScript(config) {
    const entryScript = getEntryScript(config);
    return /\.server\.[^.]+$/.test(entryScript);
}
async function resolveFile(ctx, f) {
    const index = await resolveIndexHtml(ctx);
    return path.join(path.dirname(index), f);
}
async function resolveIndexHtml(ctx) {
    const index = await ctx.resolve("/index.html");
    if (!index)
        throw new Error("Can't resolve index.html");
    return index.id;
}
function gatherCss(chunk, bundle, css) {
    function visit(chunk) {
        chunk.viteMetadata.importedCss.forEach((file) => {
            css.add(file);
        });
        chunk.imports.forEach((file) => {
            const importee = bundle[file];
            if (importee?.type === "chunk") {
                visit(importee);
            }
        });
    }
    visit(chunk);
}
function isMetaInfo(obj) {
    return !!obj && typeof obj === "object" && "meta" in obj;
}
function isWrapperInfo(obj) {
    if (isMetaInfo(obj)) {
        const { meta } = obj;
        return (!!meta &&
            typeof meta === "object" &&
            "wrapper" in meta &&
            "wrapped" in meta);
    }
    return false;
}
function readManifest(dir) {
    try {
        const f = path.join(dir, "ssr-manifest.json");
        if (fsutils.exists(f)) {
            const json = fsutils.read(f);
            fsutils.rm(f);
            const entries = Object.entries(JSON.parse(json))
                .filter(([id]) => !id.startsWith("\0"))
                .map(([id, chunks]) => [path.resolve("/", id), chunks]);
            return Object.fromEntries(entries);
        }
    }
    catch (err) {
        console.error("Failed to load ssr-manifest.json");
        console.error(err);
    }
    return {};
}
